@@ -122,6 +122,8 @@ init (CamelExchangeStore *exch, CamelExchangeStoreClass *klass)
 
 	store->flags |= CAMEL_STORE_SUBSCRIPTIONS;
 	store->flags &= ~(CAMEL_STORE_VTRASH | CAMEL_STORE_VJUNK);
+
+	exch->connect_lock = g_mutex_new ();
 }
 
 static void
@@ -136,6 +138,9 @@ finalize (CamelExchangeStore *exch)
 	
 	if (exch->folders_lock)
 		g_mutex_free (exch->folders_lock);
+
+	if (exch->connect_lock)
+		g_mutex_free (exch->connect_lock);
 }
 
 CamelType
@@ -158,6 +163,28 @@ camel_exchange_store_get_type (void)
 	return camel_exchange_store_type;
 }
 
+/* Use this to ensure that the camel session is online and we are connected
+   too. Also returns the current status of the store */
+gboolean
+camel_exchange_store_connected (CamelExchangeStore *store, CamelException *ex)
+{
+	CamelService *service;
+	CamelSession *session;
+
+	g_return_val_if_fail (CAMEL_IS_EXCHANGE_STORE (store), FALSE);
+
+	service = CAMEL_SERVICE (store);
+	session = service->session;
+
+	if (service->status != CAMEL_SERVICE_CONNECTED &&
+	    camel_session_is_online (session) &&
+	    !camel_service_connect (service, ex)) {
+		return FALSE;
+	}
+
+	return CAMEL_OFFLINE_STORE (store)->state != CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL;
+}
+
 /* This has been now removed from evolution/e-util. So implemented this here.
  * Also note that this is similar to the call in e2k-path.c. The name of the 
  * function has been changed to avoid any conflicts.
@@ -312,44 +339,118 @@ e_filename_make_safe (gchar *string)
 	}
 }
 
+static void
+camel_exchange_get_password (CamelService *service, CamelException *ex)
+{
+	CamelSession *session = camel_service_get_session (service);
+	char *prompt, *account_name;
+
+	if (!service->url->passwd) {
+		/* This is a hack to use the same string being used in 
+		   exchange-account.c:get_password */
+		account_name = g_strdup_printf ("%s@%s", service->url->user, service->url->host);
+		prompt = g_strdup_printf (_("%sEnter password for %s"), "", account_name);
+
+		service->url->passwd = camel_session_get_password (session, 
+					service, "Exchange", prompt, "password", CAMEL_SESSION_PASSWORD_SECRET, ex);
+		g_free (account_name);
+		g_free (prompt);
+	}
+}
+
+static void
+camel_exchange_forget_password (CamelService *service, CamelException *ex)
+{
+	CamelSession *session = camel_service_get_session (service);
+
+	if (service->url->passwd) {
+		camel_session_forget_password (session, 
+					       service, "Exchange", 
+					       "password", ex);
+		g_free (service->url->passwd);
+		service->url->passwd = NULL;
+	}
+}
+
 static gboolean
 exchange_connect (CamelService *service, CamelException *ex)
 {
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (service);
 	char *real_user, *socket_path;
+	gchar *password = NULL;
+	guint32 connect_status;
+	gboolean online_mode = FALSE;
 
-	if (exch->stub)
+	/* This lock is only needed for offline operation. exchange_connect
+	   is called many times in offline to ensure we are connected atleast
+	   to the mail stub. Think twice before changing anything here.*/
+	
+	g_mutex_lock (exch->connect_lock);
+
+	online_mode = camel_session_is_online (service->session);
+	
+	if (exch->stub == NULL) {
+		real_user = strpbrk (service->url->user, "\\/");
+		if (real_user)
+			real_user++;
+		else
+			real_user = service->url->user;
+		socket_path = g_strdup_printf ("/tmp/.exchange-%s/%s@%s",
+					       g_get_user_name (),
+					       real_user, service->url->host);
+		e_filename_make_safe (strchr (socket_path + 5, '/') + 1);
+		
+		exch->stub = camel_stub_new (socket_path, _("Evolution Exchange backend process"), ex);
+		g_free (socket_path);
+		if (!exch->stub) {
+			g_mutex_unlock (exch->connect_lock);
+			return FALSE;
+		}
+
+		camel_object_hook_event (CAMEL_OBJECT (exch->stub), "notification",
+					 stub_notification, exch);
+	} else if (!online_mode && exch->stub_connected) {
+		g_mutex_unlock (exch->connect_lock);
 		return TRUE;
+	}
+
+	if (online_mode) {
+		camel_exchange_get_password (service, ex);
+		if (camel_exception_is_set (ex)) {
+			g_mutex_unlock (exch->connect_lock);
+			return FALSE;
+		}
+		password = service->url->passwd;
+	}
 
-	real_user = strpbrk (service->url->user, "\\/");
-	if (real_user)
-		real_user++;
-	else
-		real_user = service->url->user;
-	socket_path = g_strdup_printf ("/tmp/.exchange-%s/%s@%s",
-				       g_get_user_name (),
-				       real_user, service->url->host);
-	e_filename_make_safe (strchr (socket_path + 5, '/') + 1);
-	
-	exch->stub = camel_stub_new (socket_path, _("Evolution Exchange backend process"), ex);
-	g_free (socket_path);
-	if (!exch->stub)
-		return FALSE;
-	
 	/* Initialize the stub connection */
 	if (!camel_stub_send (exch->stub, NULL, CAMEL_STUB_CMD_CONNECT,
+			      CAMEL_STUB_ARG_STRING, password,
 			      CAMEL_STUB_ARG_RETURN,
+			      CAMEL_STUB_ARG_UINT32, &connect_status,
 			      CAMEL_STUB_ARG_END)) {
 		/* The user cancelled the connection attempt. */
 		camel_exception_set (ex, CAMEL_EXCEPTION_USER_CANCEL,
 				     "Cancelled");
 		camel_object_unref (exch->stub);
 		exch->stub = NULL;
+		g_mutex_unlock (exch->connect_lock);
 		return FALSE;
 	}
+
+	if (!connect_status) {
+		camel_exchange_forget_password (service, ex);
+		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM,
+				     _("Could not authenticate to server. "
+				       "(Password incorrect?)\n\n"));
+		g_mutex_unlock (exch->connect_lock);
+		return FALSE;
+	} else {
+		exch->stub_connected = TRUE;
+	}
+
+	g_mutex_unlock (exch->connect_lock);
 	
-	camel_object_hook_event (CAMEL_OBJECT (exch->stub), "notification",
-				 stub_notification, exch);
 	return TRUE;
 }
 
@@ -359,8 +460,11 @@ exchange_disconnect (CamelService *service, gboolean clean, CamelException *ex)
 	return TRUE;
 }
 
-#define RETURN_VAL_IF_NOT_CONNECTED(service, ex, val)\
-	if (!exchange_connect((CamelService *)service, ex))\
+/* Even if we are disconnected, we need to exchange_connect
+   to get the offline data */
+#define RETURN_VAL_IF_NOT_CONNECTED(store, ex, val)\
+	if (!camel_exchange_store_connected(store, ex) && \
+	    !exchange_connect (CAMEL_SERVICE (store), ex)) \
 		return val; 
 
 static CamelFolder *
@@ -371,11 +475,11 @@ exchange_get_folder (CamelStore *store, const char *folder_name,
 	CamelFolder *folder;
 	char *folder_dir;
 	
-	RETURN_VAL_IF_NOT_CONNECTED(store, ex, NULL);
+	RETURN_VAL_IF_NOT_CONNECTED (exch, ex, NULL);
 
 	folder_dir = exchange_path_to_physical (exch->storage_path, folder_name);
 	
-	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+	if (!camel_exchange_store_connected (exch, ex)) {
 		if (!folder_dir || access (folder_dir, F_OK) != 0) {
 			g_free (folder_dir);
 			camel_exception_setv (ex, CAMEL_EXCEPTION_STORE_NO_FOLDER,
@@ -403,6 +507,14 @@ exchange_get_folder (CamelStore *store, const char *folder_name,
 	if (!camel_exchange_folder_construct (folder, store, folder_name,
 					      flags, folder_dir, ((CamelOfflineStore *) store)->state,
 					      exch->stub, ex)) {
+		gpointer key;
+		g_mutex_lock (exch->folders_lock);
+		if (g_hash_table_lookup_extended (exch->folders, folder_name,
+						  &key, NULL)) {
+			g_hash_table_remove (exch->folders, key);
+			g_free (key);
+		}
+		g_mutex_unlock (exch->folders_lock);
 		g_free (folder_dir);
 		camel_object_unref (CAMEL_OBJECT (folder));
 		return NULL;
@@ -430,7 +542,7 @@ exchange_folder_subscribed (CamelStore *store, const char *folder_name)
 	}
 
 	if (!camel_stub_send (exch->stub, NULL, CAMEL_STUB_CMD_IS_SUBSCRIBED_FOLDER,
-			      CAMEL_STUB_ARG_STRING, folder_name,
+			      CAMEL_STUB_ARG_FOLDER, folder_name,
 			      CAMEL_STUB_ARG_RETURN,
 			      CAMEL_STUB_ARG_UINT32, &is_subscribed,
 			      CAMEL_STUB_ARG_END)) {
@@ -447,13 +559,13 @@ exchange_subscribe_folder (CamelStore *store, const char *folder_name,
 	d(printf ("subscribe folder : %s\n", folder_name));
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (store);
 	
-	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+	if (!camel_exchange_store_connected (exch, ex)) {
 		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM, _("Cannot subscribe folder in offline mode."));
 		return;
 	}
 
 	camel_stub_send (exch->stub, ex, CAMEL_STUB_CMD_SUBSCRIBE_FOLDER,
-			      CAMEL_STUB_ARG_STRING, folder_name,
+			      CAMEL_STUB_ARG_FOLDER, folder_name,
 			      CAMEL_STUB_ARG_END);
 }
 
@@ -464,13 +576,13 @@ exchange_unsubscribe_folder (CamelStore *store, const char *folder_name,
 	d(printf ("unsubscribe folder : %s\n", folder_name));
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (store);
 	
-	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+	if (!camel_exchange_store_connected (exch, ex)) {
 		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM, _("Cannot unsubscribe folder in offline mode."));
 		return;
 	}
 
 	camel_stub_send (exch->stub, ex, CAMEL_STUB_CMD_UNSUBSCRIBE_FOLDER,
-			      CAMEL_STUB_ARG_STRING, folder_name,
+			      CAMEL_STUB_ARG_FOLDER, folder_name,
 			      CAMEL_STUB_ARG_END);
 }
 
@@ -480,7 +592,7 @@ get_trash (CamelStore *store, CamelException *ex)
 {
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (store);
 	
-	RETURN_VAL_IF_NOT_CONNECTED(store, ex, NULL);
+	RETURN_VAL_IF_NOT_CONNECTED (exch, ex, NULL);
 
 	if (!exch->trash_name) {
 		if (!camel_stub_send (exch->stub, ex, CAMEL_STUB_CMD_GET_TRASH_NAME,
@@ -501,18 +613,28 @@ make_folder_info (CamelExchangeStore *exch, char *name, char *uri,
 	d(printf ("make folder info : %s flags : %d\n", name, flags));
 	CamelFolderInfo *info;
 	const char *path;
+	gchar **components;
+	char *new_uri;
 
 	path = strstr (uri, "://");
 	if (!path)
 		return NULL;
-	path = strchr (path + 3, '/');
+	path = strstr (path + 3, "/;");
 	if (!path)
 		return NULL;
 
+	components = g_strsplit (uri, "/;", 2);
+	if (components[0] && components[1])
+		new_uri = g_strdup_printf ("%s/%s", components[0], components[1]);
+	else
+		new_uri = g_strdup (uri);
+	g_strfreev (components);
+
+	d(printf ("new_uri is : %s\n", new_uri));
 	info = g_new0 (CamelFolderInfo, 1);
 	info->name = name;
-	info->uri = uri;
-	info->full_name = g_strdup (path + 1);
+	info->uri = new_uri;
+	info->full_name = g_strdup (path + 2);
 	info->unread = unread_count;
 
 	if (flags & CAMEL_STUB_FOLDER_NOSELECT)
@@ -579,7 +701,7 @@ exchange_get_folder_info (CamelStore *store, const char *top, guint32 flags, Cam
 	 * each time auto-send/recv runs.
 	 */
 
-	RETURN_VAL_IF_NOT_CONNECTED(store, ex, NULL);
+	RETURN_VAL_IF_NOT_CONNECTED (exch, ex, NULL);
 
 	if (camel_stub_marshal_eof (exch->stub->cmd))
 		return NULL;
@@ -640,7 +762,7 @@ exchange_create_folder (CamelStore *store, const char *parent_name,
 	guint32 unread_count, flags;
 	CamelFolderInfo *info;
 	
-	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+	if (!camel_exchange_store_connected (exch, ex)) {
 		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM, _("Cannot create folder in offline mode."));
 		return NULL;
 	}
@@ -667,7 +789,7 @@ exchange_delete_folder (CamelStore *store, const char *folder_name,
 {
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (store);
 
-	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+	if (!camel_exchange_store_connected (exch, ex)) {
 		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM, _("Cannot delete folder in offline mode."));
 		return;
 	}
@@ -683,7 +805,7 @@ exchange_rename_folder (CamelStore *store, const char *old_name,
 {
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (store);
 
-	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+	if (!camel_exchange_store_connected (exch, ex)) {
 		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM, _("Cannot rename folder in offline mode."));
 		return;
 	}
@@ -794,6 +916,31 @@ stub_notification (CamelObject *object, gpointer event_data, gpointer user_data)
 		break;
 	}
 
+	case CAMEL_STUB_RETVAL_CHANGED_FLAGS_EX:
+	{
+		CamelExchangeFolder *folder;
+		char *folder_name, *uid;
+		guint32 flags;
+		guint32 mask;
+
+		if (camel_stub_marshal_decode_folder (stub->status, &folder_name) == -1 ||
+		    camel_stub_marshal_decode_string (stub->status, &uid) == -1 ||
+		    camel_stub_marshal_decode_uint32 (stub->status, &flags) == -1 ||
+		    camel_stub_marshal_decode_uint32 (stub->status, &mask) == -1)
+			break;
+
+		g_mutex_lock (exch->folders_lock);
+		folder = g_hash_table_lookup (exch->folders, folder_name);
+		g_mutex_unlock (exch->folders_lock);
+		if (folder)
+			camel_exchange_folder_update_message_flags_ex (folder, uid,
+								       flags, mask);
+
+		g_free (folder_name);
+		g_free (uid);
+		break;
+	}
+
 	case CAMEL_STUB_RETVAL_CHANGED_TAG:
 	{
 		CamelExchangeFolder *folder;
